from .TXTLoaderStrategy import TXTLoaderStrategy
from .ODTLoaderStrategy import ODTLoaderStrategy
from .DOCXLoaderStrategy import DOCXLoaderStrategy
from .CSVLoaderStrategy import CSVLoaderStrategy
from .TSVLoaderStrategy import TSVLoaderStrategy
from .XLSXLoaderStrategy import XLSXLoaderStrategy
from .ODSLoaderStrategy import ODSLoaderStrategy
from .RLoaderStrategy import RLoaderStrategy
